/** @license 2022 Neil Edelman, distributed under the terms of the
 [MIT License](https://opensource.org/licenses/MIT).

 Lexer for journal entries.

 @std C89/90 */

#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include <limits.h>
#include <errno.h>

/* This defines `enum condition`. */
/*!types:re2c*/
enum symbol { END, TEXT, BANG, WHITE, MAP };

/** scanner reads a file and extracts semantic information. Valid to access
 only while underlying pointers do not change. */
struct scanner {
	/* `re2c` variables; these point directly into `buffer`. */
	const char *marker, *ctx_marker, *from, *cursor;
	/* Weird `c2re` stuff: these fields have to come after when >5? */
	const char *label, *buffer, *s0, *s1;
	enum condition condition;
	enum symbol symbol;
	size_t line;
	int ws_before;
};

/*!re2c
re2c:yyfill:enable = 0;
re2c:flags:tags = 1;
re2c:define:YYCTYPE = char;
re2c:define:YYCURSOR = s->cursor;
re2c:define:YYMARKER = s->marker;
re2c:define:YYCTXMARKER = s->ctx_marker;
re2c:define:YYCONDTYPE = 'condition';
re2c:define:YYGETCONDITION = 's->condition';
re2c:define:YYGETCONDITION:naked = 1;
re2c:define:YYSETCONDITION = 's->condition = @@;';
re2c:define:YYSETCONDITION:naked = 1;

// Eof is marked by null when preparing files for lexing.
// Mutually exclusive; only !, [, are not covered.
end = "\x00";
newline = "\n" | "\r" "\n"?;
ws = [ \t\v\f];
glyph = [^ \t\n\r\v\f![\x00];
glyphs = glyph+;

// inside the block
decimal = [1-9][0-9]*;
number = ([1-9][0-9]* | [0])? "." [0-9]+ | [1-9][0-9]* | [0];
id = [a-zA-Z_][a-zA-Z_\-0-9]{0,63};
*/

static int lex(struct scanner *const s) {
	const char *s0, *s1;
	/*!stags:re2c format = 'const char *@@;\n'; */
	s->ws_before = 0;
scan:
	/*!re2c
	<text> end { return s->symbol = END, 1; }
	// fixme: paragraphs.
	<text> newline { s->line++; s->ws_before = 1; goto scan; }
	<text> ws+ { s->ws_before = 1; goto scan; }
	<text> @s0 glyph+ @s1 { s->s0 = s0, s->s1 = s1;
		return s->symbol = TEXT, 1; }
	<text> @s0 "!" @s1 { s->s0 = s0, s->s1 = s1;
		return s->symbol = BANG, 1; }
	<text> "![" :=> image
	<text> "[" :=> command
	<image> * { return 0; }
	<image> ws* "osm" ws* "](geo:" @s0 number "," @s1 number ")" {
		s->condition = yyctext;
		s->s0 = s0, s->s1 = s1;
		printf("Got a map.\n");
		return 1;
	}
	<command> * { return 0; }
	 */
}

static int parse_uint(const char *s, const char *e, unsigned *u) {
	uint32_t n = 0;
	for ( ; s < e; ++s) {
		unsigned digit = (unsigned)(*s - '0');
		assert(digit < 10);
		if(n > (UINT_MAX - digit) / 10) return errno = ERANGE, 0; /* check */
		n = n * 10 + digit;
	}
	*u = n;
	return 1;
}

static int parse_double(const char *s0, const char *s1, double *d) {
	char *finish;
	assert(d && s0 && s0 < s1);
	*d = strtod(s0, &finish);
	return !errno && ((finish == s1) || (errno = EDOM, 0));
}

#include <unistd.h>		/* chdir (POSIX, not ANSI) */
#include <sys/types.h>	/* mode_t (umask) */
#include <sys/stat.h>	/* umask */

int main(int argc, char **argv) {
	int success = EXIT_FAILURE;
	if(argc != 2) { fprintf(stderr, "Needs journal location.\n"
		"(should contain <year>/<month>/<day>.txt)\n"); goto finally; }
	if(chdir(argv[1]) == -1) goto catch; /* Go to journal. */

	{ success = EXIT_SUCCESS; goto finally; }
catch:
	perror("interpret");
finally:
	return EXIT_FAILURE;
}
